/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
#include "license.readme"
#include "myspAffix.h"
#include "plstr.h"
#include "nsReadableUtils.h"

/* Modifications for mozilla Copyright 2001 David Einstein Deinst@world.std.com  */

myspAffix::myspAffix(myspAffixMgr* pmgr, unsigned char a, PRInt32 num, PRInt16 ff)
{
  // initialize affix flag and register affix manager
  pmyMgr = pmgr;
  flag = a;
  
  // create and load the affentry data for this prefix
  first = new affentry[num];
  if (first) {
    numents=num;
    first->cpflag=ff;
  } else {
    /* error condition: throw exception */
  }
}

myspAffix::~myspAffix()
{
  if (first) {
    delete [] first;
  }
  first = NULL;
  numents = 0;
  flag = '\0';
  pmyMgr = NULL;
}

myspSuffix::myspSuffix(myspAffixMgr* pmgr, unsigned char a, int num, PRInt16 ff):myspAffix(pmgr,a,num,ff){}
myspPrefix::myspPrefix(myspAffixMgr* pmgr, unsigned char a, int num, PRInt16 ff ):myspAffix(pmgr,a,num,ff){}

// check if this prefix is present
const char * myspPrefix::check(const nsAFlatCString &cword)
{
  struct affentry *   pp;	// pointer to current prefix affentry
  PRInt32			cond;	// condition number being examined
  PRUint32	                tmpl;   // length of tmpword
  const char *     he;     // hash entry of root word or NULL
  nsACString::const_iterator	cp;		
  PRInt32			i;
  //char	        tmpword[MAXWORDLEN];
  nsSharableCString tmpword;

  for (pp = first, i = numents; i > 0; pp++, i--)  {

    // a match if the prefix matches the beginning of the 
    // word, if the remaining root word has positive length
    // and if there are enough chars in new root word to meet 
    // the number of characters conditions to test

    tmpl = cword.Length() - pp->appnd.Length();
    if (tmpl > 0 &&  (pp->appnd.Length() == 0 || pp->appnd.Equals(Substring(cword,0,pp->appnd.Length())))&& 
        tmpl + pp->strip.Length() >= pp->numconds) {

	    // generate new root word by removing prefix and adding
	    // back any characters that would have been stripped

      tmpword.Assign(pp->strip);
      tmpword.Append(Substring(cword,pp->appnd.Length(),cword.Length()-pp->appnd.Length()));

      // now make sure all of the conditions on characters
      // are met.  Please see the appendix at the end of
      // this file for more info on exactly what is being
      // tested

	    //cp = (unsigned char *)tmpword;
      tmpword.BeginReading(cp);
	    for (cond = 0;  cond < pp->numconds;  cond++) {
        if ((pp->conds[(unsigned char)(*cp++)] & (1 << cond)) == 0) break;
	    }

      // if all conditions are met then check if resulting
      // root word in the dictionary

	    if (cond >= pp->numconds) {
        tmpl += pp->strip.Length();
        if ((he = pmyMgr->lookup(tmpword)) != NULL) {
          if (PL_strchr(he, flag)) return he;
        }

        // prefix matched but no root word was found 
        // if XPRODUCT is allowed, try again but now 
        // ross checked combined with a suffix

        if (pp->cpflag & XPRODUCT) {
          he = pmyMgr->cross_check(tmpword, XPRODUCT, this);
          if (he) return he;
        }
	    }
    }
  }
  return NULL;
}


// see if this suffix is present in the word 
const char * myspSuffix::check(const nsAFlatCString &cword, int optflags, myspPrefix* ppfx)
{
  struct affentry *   sp;	         // suffix entry pointer
  PRInt32	                tmpl;		 // length of tmpword 
  PRInt32			cond;		 // condition beng examined
  const char *     he;              // hash entry pointer
  //unsigned char *	cp;
  nsACString::const_iterator cp;
  PRInt32			i;
  //  char	        tmpword[MAXWORDLEN];
  nsSharableCString tmpword;
  PRInt32 len=cword.Length();

  for (sp = first, i = numents; i > 0; sp++, i--) {

    // if this suffix is being cross checked with a prefix
    // but it does not support cross products skip it

    if ((optflags & XPRODUCT) != 0 &&  (sp->cpflag & XPRODUCT) == 0)
	    continue;

    // a match if the suffix matches the end of the 
    // word, if the remaining root word has positive length
    // and if there are enough chars in new root word to meet 
    // the number of characters conditions to test

    tmpl = len - sp->appnd.Length();
    if (tmpl > 0  &&  (sp->appnd.Length() == 0 || sp->appnd.Equals(Substring(cword,tmpl,sp->appnd.Length())))
        &&  tmpl + sp->strip.Length() >= sp->numconds) {

	    // generate new root word by removing suffix and adding
	    // back any characters that would have been stripped or
	    // or null terminating the shorter string
      tmpword.Assign(cword);
      tmpword.Truncate(tmpl);
	    //PL_strcpy (tmpword, word);
	    //cp = (unsigned char *)(tmpword + tmpl);
	    if (sp->strip.Length()) {
        //PL_strcpy ((char *)cp, (sp->strip.get()));
        tmpword.Append(sp->strip);
        tmpl += sp->strip.Length();
        //cp = (unsigned char *)(tmpword + tmpl);
	    } 
      //else *cp = '\0';

      // now make sure all of the conditions on characters
      // are met.  Please see the appendix at the end of
      // this file for more info on exactly what is being
      // tested
      tmpword.EndReading(cp);
	    for (cond = sp->numconds;  --cond >= 0; ) {
        if ((sp->conds[(unsigned char)(*--cp)] & (1 << cond)) == 0) break;
	    }

      // if all conditions are met then check if resulting
      // root word in the dictionary

	    if (cond < 0) {
        if ((he = pmyMgr->lookup(tmpword)) != NULL) {
          if (PL_strchr(he,flag) && 
              ((optflags & XPRODUCT) == 0 || 
               PL_strchr(he,ppfx->GetFlag()))) return he;
        }  
	    }
    }
  }
  return NULL;
}

nsresult 
myspAffix::SetAffentry(int num,const nsAFlatCString &line)
{
  PRInt32 i = 0;
  PRInt32 opos=0,pos=0;
  // split line into pieces
  struct affentry * ptr= first;
  struct affentry * nptr= first+num;
  nsACString::const_iterator begin,end;
	    
  do {
    pos=line.FindChar(' ',opos);
    if (pos != opos) {
      switch(i) 
        {
                  
          // piece 1 - is type
        case 0: 
          if (nptr != ptr) nptr->cpflag = ptr->cpflag;
          break;
                  
                  
          // piece 2 - is affix char
        case 1: 
          break;
                  
          // piece 3 - is string to strip or 0 for null 
        case 2:  
          line.BeginReading(begin);
          begin.advance(opos);
          if(pos == -1){
            line.EndReading(end);
          }
          else {
            line.BeginReading(end);
            end.advance(pos);
          }
          if(Substring(begin,end).Equals("0")){
            nptr->strip.Truncate();
          }
          else{
            nptr->strip.Adopt(ToNewCString(Substring(begin,end)));
          }
          break; 
                  
                  
          // piece 4 - is affix string or 0 for null
        case 3: 
          line.BeginReading(begin);
          begin.advance(opos);
          if(pos == -1){
            line.EndReading(end);
          }
          else {
            line.BeginReading(end);
            end.advance(pos);
          }
          if(Substring(begin,end).Equals("0")){
            nptr->appnd.Truncate();
          }
          else{
            nptr->appnd.Adopt(ToNewCString(Substring(begin,end)));
          }
          break; 
                  
                  
          // piece 5 - is the conditions descriptions
        case 4: 
          line.BeginReading(begin);
          begin.advance(opos);
          if(pos == -1){
            line.EndReading(end);
          }
          else {
            line.BeginReading(end);
            end.advance(pos);
          }
          encodeit(nptr,Substring(begin,end));
          break;

        default: 
          break;
        }
      i++;
    }
    opos=pos+1;
  }while(pos != -1);

  return NS_OK;
}


// takes aff file condition string and creates the
// conds array - please see the appendix at the end of the
// file affix.cxx which describes what is going on here
// in much more detail

void myspAffix::encodeit(struct affentry * ptr, const nsACString &code)
{
  char c;
  PRUint32 i, j, k;
  //char mbr[MAXLNLEN];
  nsSharableCString mbr;
  nsACString::const_iterator cs,d; 
  code.BeginReading(cs);
  
  // now clear the conditions array */
  for (i=0;i<SETSIZE;i++) ptr->conds[i] = (unsigned char) 0;

  // now parse the string to create the conds array */
  PRUint32 nc = code.Length();
  PRInt32 neg = 0;   // complement indicator
  PRInt32 grp = 0;   // group indicator
  PRInt32 n = 0;     // number of conditions
  PRInt32 ec = 0;    // end condition indicator
  PRInt32 nm = 0;    // number of member in group

  // if no condition just return
  if (code.Equals(".")) {
    ptr->numconds = 0;
    return;
  }

  i = 0;
  while (i < nc) {
    d = cs;
    d.advance(i);
    c = *d;

    // start group indicator
    if (c == '[') {
      grp = 1;
      c = 0;
    }

    // complement flag
    if ((grp == 1) && (c == '^')) {
      neg = 1;
      c = 0;
    }

    // end goup indicator
    if (c == ']') {
      ec = 1;
      c = 0;
    }

    // add character of group to list
    if ((grp == 1) && (c != 0)) {
      mbr.Append(c);
      //*(mbr + nm) = c;
      nm++;
      c = 0;
    }

    // end of condition 
    if (c != 0) {
      ec = 1;
    }

    
    if (ec) {
      if (grp == 1) {
        if (neg == 0) {
          // set the proper bits in the condition array vals for those chars
          for (j=0;j<mbr.Length();j++) {
            k = (PRUint32) mbr[j];
            if( k < SETSIZE){
              ptr->conds[k] = ptr->conds[k] | (1 << n);
            }
          }
        } else {
          // complement so set all of them and then unset indicated ones
          for (j=0;j<SETSIZE;j++) ptr->conds[j] = ptr->conds[j] | (1 << n);
          for (j=0;j<mbr.Length();j++) {
            k = (PRUint32) mbr[j];
            if( k < SETSIZE){
              ptr->conds[k] = ptr->conds[k] & ~(1 << n);
            }
          }
        }
        neg = 0;
        grp = 0;   
        nm = 0;
      } else {
        // not a group so just set the proper bit for this char
        ptr->conds[(unsigned int) c] = ptr->conds[(unsigned int)c] | (1 << n);
      }
      n++;
      ec = 0;
    }
    i++;
  }
  ptr->numconds = n;
  return;
}


#if 0

Appendix:  Understanding Affix Code


An affix is either a  prefix or a suffix attached to root words to make 
other words.

Basically a Prefix or a Suffix is implemented as a sequence of
affentry structures which store information about the prefix
or suffix along with supporting routines to check if a word
has a particular prefix or suffix or a combination.

The structure affentry is defined as follows:

struct affentry
{
  char * strip;          // string to strip before adding affix
  char * appnd;          // the affix string to add
  short  stripl;         // length of the strip string
  short  appndl;         // length of the affix string
  short  numconds;       // the number of conditions that must be met
  short  cpflag;         // crossproduct flag (combine both a prefix and a suffix) 
  char   conds[SETSIZE]; // array which encodes the conditions to be met
};


Here is a suffix borrowed from the en_US.aff file.  This file 
is whitespace delimited.

SFX D Y 4
SFX D   0     e          d
SFX D   y     ied        [^aeiou]y
SFX D   0     ed         [^ey]
SFX D   0     ed         [aeiou]y

This information can be interpreted as follows:

In the first line has 4 fields:

Field
-----
1     SFX - indicates this is a suffix
2     D   - is the name of the character which represents this suffix
3     Y   - indicates it can be combined with prefixes (cross product)
4     4   - indicates that sequence of 4 affentry structures are needed to
properly store the affix information

The remaining lines describe the unique information for the 4 affentry 
structures that make up this affix.  Each line can be interpreted
as follows: (note fields 1 and 2 are as a check against line 1 info)

  Field
-----
1     SFX         - indicates this is a suffix
2     D           - is the name of the character which represents this affix
3     y           - the string of chars to strip off before adding affix
                                          (a 0 here indicates the NULL string)
4     ied         - the string of affix characters to add
5     [^aeiou]y   - the conditions which must be met before the affix
can be applied

Field 5 is interesting.  Since this is a suffix, field 5 tells us that
there are 2 conditions that must be met.  The first condition is that 
the next to the last character in the word must *NOT* be any of the 
following "a", "e", "i", "o" or "u".  The second condition is that
the last character of the word must end in "y".

So how can we encode this information concisely and be able to 
test for both conditions in a fast manner?  The answer is found
but studying the wonderful ispell code of Geoff Kuenning, et.al. 
(now available under a normal BSD license).

If we set up a conds array of 256 bytes indexed (0 to 255) and access it
using a character (cast to an unsigned char) of a string, we have 8 bits
of information we can store about that character.  Specifically we
could use each bit to say if that character is allowed in any of the 
last (or first for prefixes) 8 characters of the word.

Basically, each character at one end of the word (up to the number 
of conditions) is used to index into the conds array and the resulting 
value found there says whether the that character is valid for a 
specific character position in the word.  

For prefixes, it does this by setting bit 0 if that char is valid 
in the first position, bit 1 if valid in the second position, and so on. 

If a bit is not set, then that char is not valid for that postion in the
word.

If working with suffixes bit 0 is used for the character closest 
to the front, bit 1 for the next character towards the end, ..., 
with bit numconds-1 representing the last char at the end of the string. 

Note: since entries in the conds[] are 8 bits, only 8 conditions 
read that only 8 character positions) can be examined at one
end of a word (the beginning for prefixes and the end for suffixes.

So to make this clearer, lets encode the conds array values for the 
first two affentries for the suffix D described earlier.


For the first affentry:    
numconds = 1             (only examine the last character)

conds['e'] =  (1 << 0)   (the word must end in an E)
all others are all 0

For the second affentry:
numconds = 2             (only examine the last two characters)     

conds[X] = conds[X] | (1 << 0)     (aeiou are not allowed)
where X is all characters *but* a, e, i, o, or u
         

conds['y'] = (1 << 1)     (the last char must be a y)
all other bits for all other entries in the conds array are zero


#endif

                                 
